extends Node

# Utility class for all things files
class_name FileOps

const STATS_FILE = "user://data.tres"
const STATS_PURIST_FILE = "user://purist_stats.tres"
const STATS_DAILY_FILE  = "user://daily_stats.tres"
const DAILY_FILE = "user://daily.tres"
const SETTINGS_FILE  = "user://settings.tres"
const GAMESTATE_FILE = "user://world_data.bin"
const GAMESTATE_PURIST_FILE = "user://purist_data.bin"
const GAMESTATE_DAILY_FILE  = "user://daily_data.bin"


# Moving the data file produced some issues
static func migrateLocation() -> void:
	if FileAccess.file_exists(STATS_FILE):
		var file = FileAccess.open(STATS_FILE, FileAccess.READ)
		var text = file.get_as_text()

		text = text.replace("res://Data.gd", "res://data/GameData.gd")

		var temp_file = FileAccess.open("user://data_temp.tres", FileAccess.WRITE)
		temp_file.store_string(text)
		temp_file.close()

		# Replace old file
		DirAccess.remove_absolute(STATS_FILE)
		DirAccess.rename_absolute("user://data_temp.tres", STATS_FILE)


static func saveStats(data, page = 0) -> void:
	ResourceSaver.save(data, getStatsForPage(page))


static func saveDaily(data) -> void:
	ResourceSaver.save(data, DAILY_FILE)


static func resetDailyData(today: String) -> void:
	var data = DailyData.new()
	data.lastDaily = today

	saveDaily(data)


static func loadStats(page = 0):
	if page == 0:
		if ResourceLoader.exists(STATS_FILE):
			var file = updateStatsFile(load(STATS_FILE))

			if file == null:
				migrateLocation()
				return updateStatsFile(load(STATS_FILE))
			else:
				return file
	else:
		if ResourceLoader.exists(getStatsForPage(page)):
			return updateStatsFile(load(getStatsForPage(page)))

	return GameData.new()


static func loadSettings():
	if ResourceLoader.exists(SETTINGS_FILE):
		return updateSettings(load(SETTINGS_FILE))
	return Settings.new()


static func loadDaily():
	if ResourceLoader.exists(DAILY_FILE):
		return updateDaily(load(DAILY_FILE))
	return DailyData.new()


static func getSaveForPage(page: int) -> String:
	match page:
		0: return GAMESTATE_FILE
		1: return GAMESTATE_PURIST_FILE
		2: return GAMESTATE_DAILY_FILE
		_: return ":("


static func getStatsForPage(page: int) -> String:
	match page:
		0: return STATS_FILE
		1: return STATS_PURIST_FILE
		2: return STATS_DAILY_FILE
		_: return ":("


static func deleteSave(page = 0) -> void:
	if FileAccess.file_exists(getSaveForPage(page)):
		DirAccess.remove_absolute(getSaveForPage(page))


static func saveExists(page = 0) -> bool:
	return FileAccess.file_exists(getSaveForPage(page))


static func savePage(page: int) -> void:
	var data = loadSettings()
	data.page = page
	ResourceSaver.save(data, SETTINGS_FILE)


static func lockIn(size: int) -> void:
	var data = loadDaily()
	data.lastDaily = getTodayString()
	if size == 5:
		data.playedBeginner = true
	elif size == 9:
		data.playedIntermediate = true
	else:
		data.playedMaster = true

	ResourceSaver.save(data, DAILY_FILE)


static func getTodayString() -> String:
	var today_dict = Time.get_date_dict_from_system()
	return "%04d-%02d-%02d" % [today_dict["year"], today_dict["month"], today_dict["day"]]


static func saveSettings(
	sound: bool,
	vibrate: bool,
	sensitivity,
	doubletap: bool,
	showZeros: bool,
	countDown: bool,
	showOutline: bool,
	hideSafeBlocks: bool,
	highlightZeros: bool
) -> void:
	var data = loadSettings()
	data.sound = sound
	data.vibrate = vibrate
	data.language = enumerateLang(Singleton.lang)
	data.sensitivity = sensitivity
	data.doubletap = doubletap
	data.showZeros = showZeros
	data.countDown = countDown
	data.showOutline = showOutline
	data.hideSafeBlocks = hideSafeBlocks
	data.highlightZeros = highlightZeros

	ResourceSaver.save(data, SETTINGS_FILE)


static func enumerateLang(lang) -> int:
	match lang:
		"de":
			return 1
		"esp":
			return 2
		_:
			return 0


# For when I add new things to the gamedata
static func updateStatsFile(data):
	if "version" in data and data.version == Singleton.statsVersion:
		return data
	else:
		var newData = GameData.new()
		newData.playedGames = data.playedGames
		newData.wonGames = data.wonGames
		newData.totalMines = data.totalMines
		
		newData.fastestEasy = data.fastestEasy
		newData.fastestEasyMines = data.fastestEasyMines
		newData.wonEasy = data.wonEasy
		newData.totalEasy = data.totalEasy

		newData.fastestMediumMines = data.fastestMediumMines
		newData.fastestMedium = data.fastestMedium
		newData.wonMedium = data.wonMedium
		newData.totalMedium = data.totalMedium
		
		newData.fastestHardMines = data.fastestHardMines
		newData.fastestHard = data.fastestHard
		newData.wonHard = data.wonHard
		newData.totalHard = data.totalHard

		return newData


static func updateSettings(data):
	if "version" in data and data.version == Singleton.settingsVersion:
		return data

	else:
		var newData = GameData.new()
		newData.sound = data.sound
		newData.vibrate = data.vibrate
		
		if "sensitivity" in data:
			newData.sensitivity = data.sensitivity
			newData.doubletap = data.doubletap
			newData.showZeros = data.showZeros
			newData.countDown = data.countDown

		if "showOutline" in data:
			newData.showOutline = data.showOutline
			newData.hideSafeBlocks = data.hideSafeBlocks

		if "language" in data:
			newData.language = data.language

		if "highlightZeros" in data:
			newData.highlightZeros = data.highlightZeros

		return newData


static func updateDaily(data):
	if "version" in data and data.version == Singleton.dailyVersion:
		return data

	else:
		var newData = DailyData.new()
		newData.lastDaily = data.lastDaily

		newData.playedBeginner = data.playedBeginner
		newData.playedIntermediate = data.playedIntermediate
		newData.playedMaster = data.playedMaster

		newData.timeBeginner = data.timeBeginner
		newData.timeIntermediate = data.timeIntermediate
		newData.timeMaster = data.timeMaster

		return newData
