extends Node3D

var white  = preload("res://assets/white.tres")
var orange = preload("res://assets/orange.tres")
var green  = preload("res://assets/green.tres")
var red    = preload("res://assets/red.tres")

signal exploded
signal cleared
signal defused
signal refused
signal passMine
signal clearUndoStack

var id = 0	# to make loading savefile easier

var mine = false
var hovering = false
var triggered = false
var marked = false
var dead = false
var disappearing = false
var beenBig = false
var beenFinalCleared = false
var mod = 1
var active = true

# for the last one in a disappearing island
# to save the island being gone
var saveWhenGone = false

var puristMode = Singleton.page == 1

var neighbors = []
var neighborsIds = []
var mineCount = 0
var initialMineCount = 0

var dragging = false


func _ready():
	if !mine:
		# So that loading doesn't mess up the mineCount
		if initialMineCount == 0:
			get_mine_count()
			initialMineCount = mineCount

		$MineCount.text = str(mineCount)
	else:
		$MineCount.text = ""

	update(false)


func _input(event):
	# to make sure everything is loaded before making decisions
	if get_parent().isLoading:
		return

	if active and Singleton.doubletap and !dead and !triggered and !Singleton.skip:
		if event is InputEventScreenTouch:
			if event.pressed:
				Singleton.dragging = false
			if event.double_tap:
				if hovering and !Singleton.dragging:
					if !marked:
						if Singleton.first_press:
							Singleton.first_press = false
							if mine:
								pass_mine()
						if mine:
							$BoxArea/BoxMesh.material_override = red
							if Singleton.vibrate:
								Input.vibrate_handheld(1000)
							final_clear()
							dead = true
							exploded.emit()
						else:
							dissolve(true)
					else:
						unmark()


func _process(delta):
	if get_parent().isLoading:
		return

	if active:
		if disappearing and $BoxArea/BoxMesh.scale.y > 0:
			if marked:
				$BoxArea/BoxMesh.scale += Vector3(0, 1 * delta * mod, 0)
			else:
				$BoxArea/BoxMesh.scale += Vector3(4 * delta * mod, 4 * delta * mod, 4 * delta * mod)
			if $BoxArea/BoxMesh.scale.y > 1.2 and !beenBig:
				beenBig = true
				mod *= -1
			return
		elif disappearing:
			$BoxArea.hide()
			if marked:
				for neighbor in neighbors:
					if !neighbor.mine:
						neighbor.initialMineCount -= 1
						neighbor.update(false)
				triggered = true

				if saveWhenGone:
					get_parent().saveGameBinary()

			active = false
			disappearing = false

		if !triggered and !Singleton.skip:
			if Input.is_action_just_pressed("left_mouse"):
				Singleton.dragging = false

			if !Singleton.dragging:
				if !Singleton.doubletap && Input.is_action_just_released("left_mouse"):
					if hovering and !Input.is_action_just_pressed("right_mouse"):
						if !marked:
							if Singleton.first_press:
								Singleton.first_press = false
								if mine:
									pass_mine()

							if mine:
								$BoxArea/BoxMesh.material_override = red
								if Singleton.vibrate:
									Input.vibrate_handheld(1000)
								final_clear()
								dead = true
								exploded.emit()
							else:
								# For the *real* Minesweeper experience,
								# this would have to be "false"
								dissolve(false if puristMode else true)
						else:
							unmark()

				if Input.is_action_just_pressed("right_mouse"):
					if hovering:
						if marked:
							unmark()
					
						else:
							mark()


func unmark(undoable = true) -> void:
	marked = false
	turnWhite()
	refused.emit()
	if undoable:
		get_parent().addToUndoStack(self, false)

	for neighbor in neighbors:
		if !neighbor.mine:
			neighbor.mineCount += 1
			neighbor.update()
	# Set neighbors to zero first, then check for need to decolorize boxes.
	for neighbor in neighbors:
		if !neighbor.mine:
			if Singleton.highlightZeros:
				neighbor.colorNeighbors(false)
			neighbor.update()


func mark(undoable = true) -> void:
	marked = true
	defused.emit()
	if undoable:
		get_parent().addToUndoStack(self, true)

	turnOrange()
	for neighbor in neighbors:
		if !neighbor.mine:
			neighbor.mineCount -= 1
			if Singleton.highlightZeros:
				neighbor.colorNeighbors(true)
			neighbor.update()

	check_islands()


# to reapply the color after loading the game
func reColor() -> void:
	for neighbor in neighbors:
		if !neighbor.mine:
			neighbor.colorNeighbors(true)


## Set color of all neighbors if this box was triggered before.
## If `mineCount` has just become 0, all unmarked neighbors will be turned green.
## If it has just become 1, all unmarked neighbors will be turned white.
##
## Parameters:
## -----------
## decrease: bool
##   Set to `true` if mine count has just been decreased.
func colorNeighbors(decrease: bool):
	if triggered:
		for neighbor in neighbors:
			if !neighbor.marked and !neighbor.triggered:
				await get_tree().create_timer(0.02).timeout
				if mineCount == 0 and decrease:
					neighbor.turnGreen()
				elif mineCount == 1 and !decrease:
					# Don't decolor boxes which still have "zero" neighbors.
					var hasZeroNeighbor = false
					for nextNeighbor in neighbor.neighbors:
						if (nextNeighbor.triggered and nextNeighbor.mineCount == 0):
								hasZeroNeighbor = true
								break
					if !hasZeroNeighbor:
						neighbor.turnWhite()
				neighbor.update()


# to make loading easier 
func turnOrange() -> void:
	$BoxArea/BoxMesh.material_override = orange


func turnGreen() -> void:
	$BoxArea/BoxMesh.material_override = green


func turnWhite() -> void:
	$BoxArea/BoxMesh.material_override = white


func reTrigger() -> void:
	$BoxArea.hide()
	update()


func refreshMineCount():
	mineCount = 0
	for neighbor in neighbors:
		if neighbor.mine and !neighbor.marked:
			mineCount += 1


## Return a list of neighboring boxes with mines.
func neighboringMines():
	var badNeighbors = []
	for neighbor in neighbors:
		if neighbor.mine:
			badNeighbors.append(neighbor)
	
	return badNeighbors


func is_safe():
	for neighbor in neighbors:
		if !mine || !((neighbor.triggered && neighbor.mineCount == 0) || neighbor.marked):
			return false

	return true


func check_islands():

	# When the world is being built from a file
	# this keeps it from tweaking
	var parent = get_parent()
	if !parent or !is_instance_valid(parent):
		return

	if Singleton.hideSafeBlocks and !parent.isLoading:
		var checkedNeighbors = []
		var neighborsToCheck = [self]
		if !is_safe():
			return

		while !neighborsToCheck.is_empty():
			var newNeighbors = []
			for neighbor in neighborsToCheck:
				var potentialBaddies = neighbor.neighboringMines()
				for nMine in potentialBaddies:
					if nMine not in checkedNeighbors:
						newNeighbors.append(nMine)

				checkedNeighbors.append(neighbor)

			neighborsToCheck = newNeighbors

		for cube in checkedNeighbors:
			if !cube.is_safe():
				return

		var cribSize = checkedNeighbors.size()
		var counter = 0
		for cube in checkedNeighbors:
			counter += 1
			cube.disappearing = true

			if counter == cribSize:
				cube.saveWhenGone = true

			await get_tree().create_timer(0.01).timeout
		
		# Bible says no necromancy
		clearUndoStack.emit()


func _on_box_area_mouse_entered():
	if !dead:
		hovering = true


func _on_box_area_mouse_exited():
	if !dead:
		hovering = false


func get_mine_count():
	for neighbor in neighbors:
		if neighbor.mine == true:
			mineCount += 1


func pressed():
	if marked:
		marked = false
		refused.emit()
		for neighbor in neighbors:
			if !neighbor.mine:
				neighbor.mineCount += 1
				neighbor.update()

	else:
		if !triggered:
			if Singleton.sound:
				$Plopper.play()
			cleared.emit()
		if mineCount != 0:
			$MineCount.show()
		triggered = true


func dissolve(kill: bool):
	if (!triggered and !marked and mineCount == 0 and !mine) or (!triggered and kill):
		pressed()

		for neighbor in neighbors:
			if !neighbor.triggered and !neighbor.marked and !neighbor.mine:
				await get_tree().create_timer(0.02).timeout
				neighbor.dissolve(false)
			elif neighbor.marked && neighbor.mine:
				check_islands()

	elif !mine and !marked:
		pressed()

	if !mine:
		disappearing = true

	update()


func update(check = true):
	var mc = mineCount if Singleton.countDown || (mineCount == 0 && Singleton.countDown) else initialMineCount
	$MineCount.text = str(mc)

	if mineCount == 0:
		$MineCount.set_modulate(Color.LIGHT_GRAY)
	elif mc > 4:
		$MineCount.set_modulate(Color.CRIMSON)
	elif mc > 2:
		$MineCount.set_modulate(Color.GOLD)
	elif mc > 0:
		$MineCount.set_modulate(Color.BLACK)

	if (mineCount == 0 && !Singleton.showZeros) || initialMineCount == 0:
		$MineCount.hide()
	elif triggered:
		$MineCount.show()

	if check:
		for neighbor in neighbors:
			if neighbor.marked && neighbor.mine:
				neighbor.check_islands()

## Make this box unmined.
func pass_mine():
	mine = false
	update()
	for neighbor in neighbors:
		neighbor.mineCount -= 1
		neighbor.initialMineCount -= 1
		neighbor.update()
	passMine.emit()

	refreshMineCount()
	initialMineCount = mineCount

## Make this box mined.
func recieve_mine():
	mine = true
	$MineCount.text = ""
	for neighbor in neighbors:
		neighbor.mineCount += 1
		neighbor.initialMineCount += 1
		neighbor.update()


func final_clear():
	if triggered and mineCount != 0:
		$MineCount.hide()
	if !mine:
		triggered = true
		disappearing = true
	else:
		$BoxArea/BoxMesh.material_override = red

	beenFinalCleared = true

	neighbors.shuffle()
	for neighbor in neighbors:
		if !neighbor.beenFinalCleared:
			neighbor.final_clear()
		await get_tree().create_timer(0.1).timeout


func _on_box_area_area_entered(_area):
	if !triggered:
		hide()


func _on_box_area_area_exited(_area):
	show()
